#define IMGUI_DEFINE_MATH_OPERATORS
#include <imgui_internal.h>
#include <imgui.h>
#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>

#include <glad/glad.h>
#include <GLFW/glfw3.h>
#include <glm/gtx/transform.hpp>
#include <glm/gtx/orthonormalize.hpp>

#include <iostream>
#include <memory>
#include <cmath>
#include <algorithm>

#include "Shader.h"
#include "Mesh.h"
#include "Body.h"
#include "Scene.h"
#include "Camera.h"

static const std::string ObjFile = std::string( std::string(CMAKE_SOURCE_DIR)+std::string("/models/tesseract.obj") );

/**
 * Initializes the application including the GUI.
 */
int main(int, char**)
{
    // Setup window:
    glfwSetErrorCallback([](int error, const char* description) {
        fprintf(stderr, "Glfw Error %d: %s\n", error, description);
    });

    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions:
    const char* glsl_version = "#version 330 core";
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
    // Enable AA
    glfwWindowHint(GLFW_SAMPLES, 4);
    glfwWindowHint(GLFW_MAXIMIZED, true);

    // Create window with graphics context:
    GLFWwindow* window = glfwCreateWindow(640, 640, "Computergraphik: Binary Space Partition", NULL, NULL);
    if (window == NULL)
        return 1;

    glfwMakeContextCurrent(window);
    // Enable VSync
    glfwSwapInterval(1);

    // Initialize GLAD functions:
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress)) {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    // Setup Dear ImGui context:
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    ImGui::StyleColorsClassic();

    // Setup Platform/Renderer backends:
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Load Roboto Font:
    io.Fonts->AddFontFromFileTTF(CMAKE_SOURCE_DIR "/lib/imgui-1.88/misc/fonts/Roboto-Medium.ttf", 24.0f);

    // Enable depth test:
    glEnable(GL_DEPTH_TEST);

    // Enable anti-aliasing
    glEnable(GL_MULTISAMPLE);

    // *** Create our scene ***
    auto scene = std::make_shared<Scene>();

    // Placing a table at the origin
    auto tesseract = std::make_shared<Body>( std::vector<const char*>{ObjFile.c_str()}, false );
    tesseract->setScale( glm::vec3(0.01) );
    tesseract->setRotation( glm::rotate((float)(M_PI/4.f), glm::normalize(glm::vec3(1,0,0))) * glm::rotate((float)(M_PI/2.f), glm::normalize(glm::vec3(0,0,1))) );
    tesseract->getMeshAt(0)->m_color = glm::vec4(0,0.310,0.235,1)*1.2f;
    tesseract->getMeshAt(0)->m_shininess = 10.f;
    tesseract->getMeshAt(0)->m_specularStrength = 0.5f;
    tesseract->simulateDynamics = false;
    scene->bodies.push_back( tesseract );

    // Create BSP-Tree
    tesseract->getMeshAt(0)->updateBsp();

    // Start simulation
    srand(time(NULL));
    Camera camera;

    // Reference old mouse position to determine change in position => dragging
    glm::vec3 lastTrackballPos(0,0,0);
    glm::vec3 eye(0,0,5);
    glm::mat3 M_trackball = glm::identity<glm::mat3>();
    glm::mat4 view = glm::lookAt(eye, glm::vec3(0,0,0), glm::vec3(0,1,0));
    bool leftMouseUpBefore = true;
    bool spaceUpBefore = true;

    // We use a simple phong shader for all geometry
    Shader shader(CMAKE_SOURCE_DIR "/shader/phong.vs", CMAKE_SOURCE_DIR "/shader/phong.fs");

    // Transparent rendering
    glDisable( GL_DEPTH_TEST );
	glDisable( GL_CULL_FACE );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

    // Rendering modes
    int currentMode = 0;
    const char* modeNames[] = {"Unsortiert", "BSP (ohne Splitting)", "BSP (mit Splitting)"};

    // Main loop which is executed every frame until the window is closed
    while (!glfwWindowShouldClose(window))
    {
        // Processes all glfw events:
        glfwPollEvents();

        // Get the size of the window:
        int display_w, display_h;
        glfwGetFramebufferSize(window, &display_w, &display_h);

        // Create the GUI:
        {
            // Start the Dear ImGui frame
            ImGui_ImplOpenGL3_NewFrame();
            ImGui_ImplGlfw_NewFrame();
            ImGui::NewFrame();
        }

        // Setup the GL viewport
        glViewport(0, 0, display_w, display_h);
        glClearColor(1.f, 1.f, 1.f, 1.f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Calculate aspect ratio (might be used for perspective transformation):
        float aspectRatio = float(display_w) / display_h;

        M_trackball = glm::identity<glm::mat4>();

        // Interactive controls
        bool leftMouseDown = ImGui::IsMouseDown(ImGuiMouseButton_Left);
        bool rightMouseDown = ImGui::IsMouseDown(ImGuiMouseButton_Right);
        bool spaceDown = ImGui::IsKeyDown(ImGuiKey_Space);
        if (!ImGui::GetIO().WantCaptureMouse) {
            auto m = glm::vec2(ImGui::GetMousePos().x, ImGui::GetMousePos().y);
            float w = display_w;
            float h = display_h;
#if defined(__APPLE__)
            h /= 2;
            w /= 2;
#endif
            auto trackballPos_screen = Camera::ScreenToTrackball(m, glm::vec2(w,h));
            auto trackballPos = trackballPos_screen;

            // Left mouse button changed from up to down?
            if (leftMouseDown) {
                if (leftMouseUpBefore)
                    lastTrackballPos = trackballPos;
                leftMouseUpBefore = false;
            } else
                leftMouseUpBefore = true;

            // Rotate trackball
            if (leftMouseDown) {
                auto r2 = glm::vec4(glm::cross(lastTrackballPos, trackballPos),0);
                float a2 = acosf(glm::dot(glm::normalize(lastTrackballPos),glm::normalize(trackballPos))) * 1.f;
                if (glm::length(r2)>0.0001f && fabsf(a2)>0.0001f) {
                    M_trackball = glm::rotate(a2,glm::vec3(r2));
                    lastTrackballPos = trackballPos;
                }
            }
        }

        // Rotate view matrix
        view = glm::translate(-eye) * glm::mat4(M_trackball) * glm::translate(eye) * view;

        // Update shader state
        static constexpr float Fov = 25; // Low FOV for less perspective distortion
        glm::mat4 projection = glm::perspective( glm::radians(Fov), aspectRatio, 0.01f, 1000.f);
        shader.setUniform("projection", projection);
        shader.setUniform("view", view);
        // Extracting the view direction
        glm::vec3 center_eye( view[0][2], view[1][2], view[2][2] );
        // Extracting the view distance
        float center_eye_dot_eye = view[3][2];
        // Technically this is not 100% correct, but the error is rather low
        // FIXME: The correct value can be retrieved from the Camera class
        glm::vec3 view_pos = -center_eye*center_eye_dot_eye;
        shader.setUniform("viewPos", view_pos);
        shader.setUniform("lightColor", glm::vec3(0.925f,0.9f,1.0f));
        shader.setUniform("lightPos", glm::vec3(0,0,5.f));

        // Debug window
        static constexpr float DebugWindowWidth = 500;
        static constexpr float DebugWindowHeight = 4*40;
        float adjustedHeight = display_h-DebugWindowHeight;
#if defined(__APPLE__)
        adjustedHeight = display_h/2.f-DebugWindowHeight;
#endif
        ImGui::SetNextWindowPos(ImVec2(0,0));
        ImGui::SetNextWindowSize(ImVec2(DebugWindowWidth,DebugWindowHeight));
        if (ImGui::Begin("##empty", nullptr, ImGuiWindowFlags_NoCollapse|ImGuiWindowFlags_NoResize)) {
            ImGui::Combo("Sortierung der Dreiecke", &currentMode, &modeNames[0], IM_ARRAYSIZE(modeNames));
            ImGui::Text("FPS: %f\n", ImGui::GetIO().Framerate);
            ImGui::Text("Camera Position: %f,%f,%f\n", view_pos.x, view_pos.y, view_pos.z);
            ImGui::Text("Triangle Count: %i\n", tesseract->getMeshAt(0)->m_triangleCount);
            ImGui::End();
        }
        tesseract->getMeshAt(0)->m_renderMode = (RenderMode)currentMode;

        // Render all meshes
        for (auto body : scene->bodies)
            body->render( shader, glm::vec3(glm::inverse(tesseract->getModelToWorld()) * glm::vec4(view_pos,1)) );

        // Render the GUI and draw it to the screen:
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap Buffers:
        glfwSwapBuffers(window);
    }

    // Cleanup
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(window);
    glfwTerminate();

    return 0;
}